import datetime
import persistent
from zope import interface
from zope import schema
import zope.interface.interface

import collective.singing.interfaces
from collective.singing import MessageFactory as _

class ISubjectsCollectorBase(collective.singing.interfaces.ICollector):
    pass

class IFilteredSubjectsCollectorBase(collective.singing.interfaces.ICollector):
    pass

class SubjectsCollectorBase(persistent.Persistent):
    """A template class that allows you to create a simple collector
    that presents one field with a vocabulary to the user.

    You can provide the vocabulary and the title of the field by
    overriding methods and attributes.
    """
    interface.implements(ISubjectsCollectorBase)

    title = _(u"Subjects collector")

    field_name = 'subjects'
    field_title = _(u"Subjects")

    def __init__(self, id, title):
        self.id = id
        self.title = title
        super(SubjectsCollectorBase, self).__init__()

    @property
    def full_schema(self):
        vocabulary = self.vocabulary()
        field = schema.Set(
            __name__=self.field_name,
            title=self.field_title,
            value_type=schema.Choice(vocabulary=vocabulary))

        interface.directlyProvides(
            field, collective.singing.interfaces.IDynamicVocabularyCollection)

        return zope.interface.interface.InterfaceClass(
            'Schema',
            bases=(collective.singing.interfaces.ICollectorSchema,),
            attrs={field.__name__: field})

    @property
    def schema(self):
        vocabulary = self._vocabulary()
        field = schema.Set(
            __name__=self.field_name,
            title=self.field_title,
            value_type=schema.Choice(vocabulary=vocabulary))

        interface.directlyProvides(
            field, collective.singing.interfaces.IDynamicVocabularyCollection)

        return zope.interface.interface.InterfaceClass(
            'Schema',
            bases=(collective.singing.interfaces.ICollectorSchema,),
            attrs={field.__name__: field})

    def get_items(self, cue=None, subscription=None):
        if subscription is not None:
            data = subscription.collector_data.get(
                self.field_name, set())
        else:
            data = set()

        return self.get_items_for_selection(cue, data), self.now()

    def now(self):
        return datetime.datetime.now()
    
    def _vocabulary(self):
        return self.vocabulary()

    def get_items_for_selection(self, cue, data):
        """Override this method and return a list of items that match
        the set of choices from the vocabulary given in ``data``.  Do
        not return items that are older than ``cue``.
        """
        raise NotImplementedError()

    def vocabulary(self):
        """Override this method and return a zope.schema.vocabulary
        vocabulary.
        """
        raise NotImplementedError()

class FilteredSubjectsCollectorBase(SubjectsCollectorBase):
    interface.implements(IFilteredSubjectsCollectorBase)

    filtered_items = []

    def _vocabulary(self):
        if self.filtered_items:
            return zope.schema.vocabulary.SimpleVocabulary(
                [t for t in self.vocabulary()
                 if t.token in self.filtered_items])
        else:
            return self.vocabulary()




# So this is a super-ugly hack to try and satisfy the zope.app.catalog
# dependency to a newer version of BTrees.

import BTrees
#import BTrees.Interfaces

class _Family(object):
    #zope.interface.implements(BTrees.Interfaces.IBTreeFamily)

    from BTrees import OOBTree as OO

class _Family32(_Family):
    from BTrees import OIBTree as OI
    from BTrees import IIBTree as II
    from BTrees import IOBTree as IO
    from BTrees import IFBTree as IF

    maxint = int(2**31-1)
    minint = -maxint - 1

    def __reduce__(self):
        return _family32, ()

# class _Family64(_Family):
#     from BTrees import OLBTree as OI
#     from BTrees import LLBTree as II
#     from BTrees import LOBTree as IO
#     from BTrees import LFBTree as IF

#     maxint = 2**63-1
#     minint = -maxint - 1

#     def __reduce__(self):
#         return _family64, ()

def _family32():
    return family32
_family32.__safe_for_unpickling__ = True

# def _family64():
#     return family64
# _family64.__safe_for_unpickling__ = True


BTrees.family32 = family32 = _Family32()
# family64 = _Family64()


# BTrees.family64.IO.family = family64
# BTrees.family64.OI.family = family64
# BTrees.family64.IF.family = family64
# BTrees.family64.II.family = family64

BTrees.family32.IO.family = family32
BTrees.family32.OI.family = family32
BTrees.family32.IF.family = family32
BTrees.family32.II.family = family32



"""A simplistic wizard for z3c.form that keeps previously entered
values in hidden form elements.
"""

from zope import schema
from zope.app.pagetemplate import viewpagetemplatefile
from z3c.form import button, field, form
import z3c.form.interfaces

from collective.singing import MessageFactory as _
from collective.singing.browser import utils

class Step(utils.OverridableTemplate, form.Form):
    index = viewpagetemplatefile.ViewPageTemplateFile('wizard-step.pt')
    subforms = ()
    label = u""
    description = u""
    ignoreContext = True

    def __init__(self, context, request, parent):
        super(Step, self).__init__(context, request)
        self.parent = parent

class Wizard(utils.OverridableTemplate, form.Form):

    success_message = _(u"Information submitted successfully.")
    errors_message = _(u"There were errors.")

    index = viewpagetemplatefile.ViewPageTemplateFile('wizard.pt')
    finished = False
    steps = () # Set this to be form classes
    label = u""
    description = u""
    ignoreContext = True

    fields = field.Fields(schema.Int(__name__='step', default=-1))

    def update(self):
        self.updateWidgets()
        self.before_steps = []
        self.current_step = None
        self.current_index = int(self.widgets['step'].value)

        # Don't attempt to extract from the current step if we're
        # viewing the form for the first time
        boot = False
        if self.current_index == -1:
            self.current_index = 0
            boot = True

        self.update_steps()
        
        # If we're viewing the form for the first time, let's set the
        # step to 0
        if boot:
            self.widgets['step'].value = str(0)

        # Hide all widgets from previous steps
        self._hide_widgets()
        self.widgets['step'].mode = z3c.form.interfaces.HIDDEN_MODE

        self.updateActions()
        self.actions.execute()

    def _hide_widgets(self):
        for step in self.before_steps:
            for widget in step.widgets.values():
                widget.mode = z3c.form.interfaces.HIDDEN_MODE

    def update_steps(self):
        self.before_steps = []
        for index in range(self.current_index):
            step = self.steps[index](self.context, self.request, self)
            step.update()
            self.before_steps.append(step)
            
        self.current_step = self.steps[self.current_index](
            self.context, self.request, self)
        self.current_step.update()

    def is_last_step(self):
        return len(self.before_steps) == len(self.steps) - 1

    @button.buttonAndHandler(_(u'Proceed'),
                             name='proceed',
                             condition=lambda form:not form.is_last_step())
    def handle_proceed(self, action):
        data, errors = self.current_step.extractData()
        if errors:
            self.status = self.errors_message
        else:
            self.current_index = current_index = self.current_index + 1
            self.widgets['step'].value = current_index
            self.update_steps()
            self._hide_widgets()

            # Proceed can change the conditions for the finish button,
            # so we need to reconstruct the button actions, since we
            # do not redirect.
            self.updateActions()

    @button.buttonAndHandler(_(u'Finish'),
                             name='finish',
                             condition=lambda form:form.is_last_step())
    def handle_finish(self, action):
        data, errors = self.current_step.extractData()
        if errors:
            self.status = self.errors_message
            return
        else:
            self.status = self.success_message
            self.finished = True

        data, errors = self.extractData()
        self.finish(data)

    def extractData(self):
        steps = self.before_steps + [self.current_step]
        return utils.extract_data_prefixed(steps)

    def finish(self, data):
        raise NotImplementedError



import datetime

import persistent
import persistent.list
from zope import interface
from zope import component
from zope.app.component.hooks import getSiteManager

class IQueue(interface.Interface):
    pass

class Queue(persistent.Persistent):
    interface.implements(IQueue)

    def __init__(self):
        self.pending = persistent.list.PersistentList()
        self.finished = persistent.list.PersistentList()

    def process(self):
        num = len(self.pending)
        for job in self.pending:
            job()
            self.pending.remove(job)
            self.finished.append(job)
        return num

class Job(persistent.Persistent):
    executed = None
    title = u''

    def __init__(self, fun, *args, **kwargs):
        self._fun = fun
        self._args = args
        self._kwargs = kwargs

    def __call__(self):
        self.value = self._fun(*self._args, **self._kwargs)
        self.executed = datetime.datetime.now()

def get_queue(name):
    queue = component.queryUtility(IQueue, name)
    if queue is None:
        queue = Queue()
        sm = getSiteManager()
        sm.registerUtility(queue, provided=IQueue, name=name)
    return queue



import pprint

import persistent
import persistent.dict
import persistent.list
from zope import interface
from zope import component
import zope.event
import zope.lifecycleevent
import zope.index.text.interfaces
import zope.app.catalog.catalog
import zope.app.catalog.field
import zope.app.catalog.text
import zope.app.intid.interfaces
import zope.app.container.btree

from collective.singing import interfaces
import collective.singing.subscribe
from collective.singing import MessageFactory as _

def secret(channel, composer, data, request):
    """Look up an appropriate secret.

    Looks for a IRequestBasedSecret first, and falls back to an
    IComposerBasedSecret if none is available or if the
    IRequestBasedSecret component returns None.
    """
    try:
        rbs = component.getUtility(interfaces.IRequestBasedSecret)
    except component.ComponentLookupError, e:
        pass
    else:
        secret = rbs(request)
        if secret is not None:
            return secret
    
    cbs = interfaces.IComposerBasedSecret(composer)
    return cbs.secret(data)

def has_single_format(channel):
    """Returns format of a given channel's single composer, if the
    channel has only one composer; otherwise returns None.
    
      >>> class channel:
      ...     composers = {'sms': object()}
      >>> has_single_format(channel) == 'sms'
      True

      >>> channel.composers = {}
      >>> has_single_format(channel) # doctest: +ELLIPSIS
      Traceback (most recent call last):
      ...
      AssertionError...

      >>> channel.composers = {'sms': None, 'email': None}
      >>> has_single_format(channel) is None
      True
    """
    composers = channel.composers.keys()
    assert len(composers) >= 1

    if len(composers) == 1:
        return composers[0]

class SimpleSubscription(persistent.Persistent):
    """
      >>> from zope.interface.verify import verifyClass
      >>> verifyClass(interfaces.ISubscription, SimpleSubscription)
      True
    """
    interface.implements(interfaces.ISubscription)

    def __init__(self, channel, secret,
                 composer_data, collector_data, metadata):
        self.channel = channel
        self.secret = secret

        self.composer_data = persistent.dict.PersistentDict()
        self.collector_data = persistent.dict.PersistentDict()
        self.metadata = persistent.dict.PersistentDict()
        self.composer_data.update(composer_data)
        self.collector_data.update(collector_data)
        self.metadata.update(metadata)

    def __repr__(self):
        def dict_format(data):
            return pprint.pformat(dict(data)).replace('\n', '')

        data = dict(channel=self.channel)
        for attr in ('composer_data', 'collector_data', 'metadata'):
            data[attr] = dict_format(getattr(self, attr))

        fmt_str = "<SimpleSubscription to %(channel)r with composerdata: %(composer_data)s, collectordata: %(collector_data)s, and metadata: %(metadata)s>"
        return fmt_str % data

class ISubscriptionCatalogData(interface.Interface):
    """Extract metadata from subscription for use in catalog.
    """
    format = interface.Attribute(u"Format")
    key = interface.Attribute(u"Key")
    label = interface.Attribute(u"Label")

@interface.implementer(ISubscriptionCatalogData)
@component.adapter(interfaces.ISubscription)
def catalog_data(subscription):
    class Metadata(object):
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

    metadata = subscription.metadata
    composer_data = subscription.composer_data
    cschema = subscription.channel.composers[metadata['format']].schema

    def _find_field(schema, interface):
        for name in schema:
            if interface.providedBy(cschema[name]):
                return cschema[name]
        else:
            return cschema[cschema.names()[0]]

    key_field = _find_field(cschema, interfaces.ISubscriptionKey)
    label_field = _find_field(cschema, interfaces.ISubscriptionLabel)

    return Metadata(format=metadata['format'],
                    key=composer_data[key_field.__name__],
                    label=composer_data[label_field.__name__])

class SubscriptionSearchableText(object):
    component.adapts(interfaces.ISubscription)
    interface.implements(zope.index.text.interfaces.ISearchableText)

    def __init__(self, subscription):
        self.subscription = subscription

    def getSearchableText(self):
        return u' '.join(
            unicode(v) for v in self.subscription.composer_data.values())

class Subscriptions(zope.app.container.btree.BTreeContainer):
    """An ISubscriptions implementation that's based on ZODB and uses
    a zope.app.catalog Catalog to provide the query interface.

      >>> from zope.interface.verify import verifyObject
      >>> verifyObject(interfaces.ISubscriptions, Subscriptions())
      True
    """
    interface.implements(interfaces.ISubscriptions)

    subscription_factory = SimpleSubscription

    def __init__(self):
        super(Subscriptions, self).__init__()
        self._catalog = zope.app.catalog.catalog.Catalog()

        fulltext = zope.app.catalog.text.TextIndex(
            interface=zope.index.text.interfaces.ISearchableText,
            field_name='getSearchableText',
            field_callable=True)

        secret = zope.app.catalog.field.FieldIndex(
            interface=interfaces.ISubscription,
            field_name='secret',
            )

        format = zope.app.catalog.field.FieldIndex(
            interface=ISubscriptionCatalogData,
            field_name='format',
            )

        key = zope.app.catalog.field.FieldIndex(
            interface=ISubscriptionCatalogData,
            field_name='key',
            )

        label = zope.app.catalog.text.TextIndex(
            interface=ISubscriptionCatalogData,
            field_name='label',
            )

        self._catalog[u'fulltext'] = fulltext
        self._catalog[u'secret'] = secret
        self._catalog[u'format'] = format
        self._catalog[u'key'] = key
        self._catalog[u'label'] = label

    def query(self, **kwargs):
        query = {}
        for key, value in kwargs.items():
            index = self._catalog[key]
            if isinstance(index, zope.app.catalog.field.FieldIndex):
                value = (value, value)
            query[key] = value
        return self._catalog.searchResults(**query)

    def add_subscription(
        self, channel, secret, composerd, collectord, metadata):
        subscription = self.subscription_factory(
            channel, secret, composerd, collectord, metadata)

        data = ISubscriptionCatalogData(subscription)
        contained_name = u'%s-%s' % (data.key, data.format)
        if contained_name in self:
            raise ValueError(_("There's already a subscription for ${name}",
                               mapping=dict(name=contained_name)))
        self[contained_name] = subscription
        return self[contained_name]

    def remove_subscription(self, subscription):
        data = ISubscriptionCatalogData(subscription)
        del self[u'%s-%s' % (data.key, data.format)]

def _catalog_subscription(subscription):
    intids = component.getUtility(zope.app.intid.interfaces.IIntIds)
    subscription.channel.subscriptions._catalog.index_doc(
        intids.getId(subscription), subscription)

@component.adapter(collective.singing.interfaces.ISubscription,
                   zope.app.container.interfaces.IObjectAddedEvent)
def subscription_added(obj, event):
    intids = component.getUtility(zope.app.intid.interfaces.IIntIds)
    intids.register(obj)
    _catalog_subscription(obj)

@component.adapter(collective.singing.interfaces.ISubscription,
                   zope.lifecycleevent.IObjectModifiedEvent)
def subscription_modified(obj, event):
    _catalog_subscription(obj)

@component.adapter(collective.singing.interfaces.ISubscription,
                   zope.app.intid.interfaces.IIntIdRemovedEvent)
def subscription_removed(obj, event):
    intids = component.getUtility(zope.app.intid.interfaces.IIntIds)
    obj.channel.subscriptions._catalog.unindex_doc(intids.getId(obj))



from zope import interface
from zope import component
import zope.schema.interfaces
import zope.schema.vocabulary

from collective.singing import interfaces
from collective.singing.channel import channel_lookup

def channel_vocabulary(context):
    terms = []
    for channel in channel_lookup():
        terms.append(
            zope.schema.vocabulary.SimpleTerm(
                value=channel,
                token=channel.name,
                title=channel.title))
    return zope.schema.vocabulary.SimpleVocabulary(terms)
interface.alsoProvides(channel_vocabulary,
                       zope.schema.interfaces.IVocabularyFactory)

def subscribeable_channel_vocabulary(context):
    terms = []
    for channel in channel_lookup(only_subscribeable=True):
        terms.append(
            zope.schema.vocabulary.SimpleTerm(
                value=channel,
                token=channel.name,
                title=channel.title))
    return zope.schema.vocabulary.SimpleVocabulary(terms)
zope.interface.alsoProvides(subscribeable_channel_vocabulary,
                            zope.schema.interfaces.IVocabularyFactory)

def sendable_channel_vocabulary(context):
    terms = []
    for channel in channel_lookup(only_sendable=True):
        terms.append(
            zope.schema.vocabulary.SimpleTerm(
                value=channel,
                token=channel.name,
                title=channel.title))
    return zope.schema.vocabulary.SimpleVocabulary(terms)
zope.interface.alsoProvides(sendable_channel_vocabulary,
                            zope.schema.interfaces.IVocabularyFactory)



import os
from setuptools import setup, find_packages

version = '0.6.7'

def read(*pathnames):
    return open(os.path.join(os.path.dirname(__file__), *pathnames)).read()

setup(name='collective.singing',
      version=version,
      description="A Zope 3 library for sending notifications and newsletters",
      long_description='\n'.join([
          read('docs', 'README.txt'),
          read('docs', 'HISTORY.txt'),
          ]),
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='zope plone notification newsletter',
      author='Daniel Nouri, Thomas Clement Mogensen and contributors',
      author_email='singing-dancing@googlegroups.com',
      url='http://pypi.python.org/pypi/collective.singing',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
          'setuptools',
          'zope.sendmail',
          'zope.app.catalog',
          'zc.queue',
          'zc.lockfile',
          'plone.z3cform>=0.5.1',
      ],
      
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
